from collections import OrderedDict, Sequence
import toolz as t
import toolz.curried as tc
from boltons import funcutils as bfu
from joblib.func_inspect import getfullargspec

from provenance.compatibility import getargspec


UNSPECIFIED_ARG = '::unspecified::'


def args_extractor(f, merge_defaults=False):
    """
    Takes a function, inspects it's parameter lists, and returns a
    function that will return all of the named and key arguments
    back as a dictionary. The varargs are also returned which don't
    have a names.

    """
    spec = getfullargspec(f)
    if spec.defaults:
        param_defaults = dict(zip(spec.args[-len(spec.defaults):],
                                   spec.defaults))
    else:
        param_defaults = {}
    named_param_defaults = spec.kwonlydefaults or {}
    default_dicts = {}
    num_named_args = len(spec.args)

    if merge_defaults is True and hasattr(f, '__merge_defaults__'):
        merge_defaults = f.__merge_defaults__


    if merge_defaults:
        default_dicts = t.pipe(t.merge(named_param_defaults, param_defaults),
                               tc.valfilter(lambda v: isinstance(v, dict)))

        if isinstance(merge_defaults, Sequence):
            default_dicts = {k:default_dicts[k] for k in merge_defaults}

        def _args_dict(args, kargs):
            unnamed_args = dict(zip(spec.args, args[0:num_named_args]))
            varargs = args[num_named_args:]
            kargs = t.merge(kargs, unnamed_args)
            for k, d in default_dicts.items():
                kargs[k] = t.merge(d, kargs.get(k) or {})
            return varargs, kargs
    else:
        def _args_dict(args, kargs):
            unnamed_args = dict(zip(spec.args, args[0:num_named_args]))
            varargs = args[num_named_args:]
            kargs = t.merge(kargs, unnamed_args)
            return varargs, kargs

    return _args_dict


def with_merged_defaults(*kwargs_to_default):
    """
    Introspects the argspec of the function being decorated to see what
    keyword arguments take dictionaries. If a dictionary is passed in when
    then function is called then it is merged with the dictionary defined
    in the parameter list.
    """
    merge_defaults = True
    if len(kwargs_to_default) > 0:
        merge_defaults = kwargs_to_default

    def _with_merged_defaults(f):
        extract_kargs = args_extractor(f, merge_defaults)

        @bfu.wraps(f)
        def _merge_defaults(*args, **kargs):
            vargs, kargs = extract_kargs(args, kargs)
            return f(*vargs, **kargs)
        _merge_defaults.__merge_defaults__ = merge_defaults

        return _merge_defaults
    return _with_merged_defaults


def is_curry_func(f):
    """
    Checks if f is a toolz or cytoolz function by inspecting the available attributes.
    Avoids explicit type checking to accommodate all versions of the curry fn.
    """
    return hasattr(f, 'func') and hasattr(f, 'args') and hasattr(f, 'keywords')


def _func_param_info(argspec):
    params = argspec.args
    defaults = argspec.defaults or []
    start_default_ix = -max(len(defaults), 1) - 1
    values = [UNSPECIFIED_ARG] * (len(params) - len(defaults)) + \
             list(defaults[start_default_ix:])
    return OrderedDict(zip(params, values))


def param_info(f):
    if is_curry_func(f):
        argspec = getargspec(f.func)
        num_args = len(f.args)
        args_to_remove = argspec.args[0:num_args] + list(f.keywords.keys())
        base = _func_param_info(argspec)
        return t.dissoc(base, *args_to_remove)
    return(_func_param_info(getargspec(f)))


def inner_function(partial_fn):
    """Returns the wrapped function of either a partial or curried function."""
    fn = partial_fn.func
    if '__module__' not in dir(fn):
        # for some reason the curry decorator nests the actual function
        # metadata one level deeper
        fn = fn.func
    return fn


def partial_fn_info(partial_fn):
    fn = inner_function(partial_fn)
    varargs, kargs = args_extractor(fn)(partial_fn.args, partial_fn.keywords)
    return {'varargs': varargs, 'kargs': kargs,
            'module': fn.__module__, 'name': fn.__name__}

# TODO: consider using the functions in joblib.func_inspect, e.g. for the fn name
def fn_info(fn):
    if 'func' in dir(fn):
        return partial_fn_info(fn)
    return {'name': fn.__name__, 'module': fn.__module__,
            'varargs': (), 'kargs': {}}

def when_type(type):
    def _decorator(f):
        @bfu.wraps(f)
        def _when_type(val):
            if isinstance(val, type):
                return f(val)
            else:
                return val
        return _when_type
    return _decorator



from __future__ import absolute_import, division, print_function

from ._config import from_config, load_config, load_yaml_config
from ._dependencies import dependencies
from .core import provenance, provenance_set, promote, archive_file
from .repos import (capture_set, create_set, get_default_repo, set_run_info_fn,
                    get_set_by_id, get_set_by_name, lazy_dict, lazy_proxy_dict,
                    load_artifact, load_proxy, set_default_repo, using_repo)

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



import copy
from datetime import datetime

import sqlalchemy as sa
import sqlalchemy.dialects.postgresql as pg
import sqlalchemy.ext.declarative
import sqlalchemy.orm
from memoized_property import memoized_property

Base = sa.ext.declarative.declarative_base()

SHA1_LENGTH = 40
VALUE_ID_LENGTH = SHA1_LENGTH + 10 # extra 10 for optional file extension info

class Run(Base):
    __tablename__ = 'runs'

    id = sa.Column(pg.VARCHAR(SHA1_LENGTH), primary_key=True)
    hostname = sa.Column(pg.VARCHAR(256))
    info = sa.Column(pg.JSONB)
    created_at = sa.Column(pg.TIMESTAMP, default=datetime.utcnow)
    artifacts = sqlalchemy.orm.relationship("Artifact")

    def __init__(self, info):
        self.id = info['id']
        self.info = info
        self.hostname = info['host']['nodename']
        self.created_at = info['created_at']

    @memoized_property
    def info_with_datetimes(self):
        result = copy.copy(self.info)
        result['created_at'] = self.created_at
        return result

class Artifact(Base):
    __tablename__ = 'artifacts'

    id = sa.Column(pg.VARCHAR(SHA1_LENGTH), primary_key=True)
    value_id = sa.Column(pg.VARCHAR(VALUE_ID_LENGTH))
    run_id = sa.Column(pg.VARCHAR(SHA1_LENGTH), sa.ForeignKey("runs.id"))

    name = sa.Column(pg.VARCHAR(1000))
    version = sa.Column(pg.INTEGER)
    fn_module = sa.Column(pg.VARCHAR(100))
    fn_name = sa.Column(pg.VARCHAR(100))

    composite = sa.Column(pg.BOOLEAN)

    value_id_duration = sa.Column(pg.FLOAT)
    compute_duration = sa.Column(pg.FLOAT)
    hash_duration = sa.Column(pg.FLOAT)

    computed_at = sa.Column(pg.TIMESTAMP)
    added_at = sa.Column(pg.TIMESTAMP, default=datetime.utcnow)

    host = sa.Column(pg.JSONB)
    process = sa.Column(pg.JSONB)
    input_artifact_ids = sa.Column(pg.ARRAY(pg.VARCHAR(SHA1_LENGTH)))
    inputs_json = sa.orm.deferred(sa.Column(pg.JSONB))
    serializer = sa.Column(pg.VARCHAR(128), default='joblib')
    load_kwargs = sa.Column(pg.JSONB)
    dump_kwargs = sa.Column(pg.JSONB)
    custom_fields = sa.Column(pg.JSONB)

    def __init__(self, artifact, inputs_json, run):
        self.id = artifact.id
        self.run = run
        self.run_id = run.id
        self.value_id = artifact.value_id
        self.name = artifact.name
        self.version = artifact.version
        self.fn_module = artifact.fn_module
        self.fn_name = artifact.fn_name
        self.composite = artifact.composite
        self.value_id_duration = artifact.value_id_duration
        self.compute_duration = artifact.compute_duration
        self.hash_duration = artifact.hash_duration
        self.input_artifact_ids = artifact.input_artifact_ids
        self.inputs_json = inputs_json
        self.custom_fields = artifact.custom_fields
        self.computed_at = artifact.computed_at
        self.serializer = artifact.serializer
        self.load_kwargs = artifact.load_kwargs
        self.dump_kwargs = artifact.dump_kwargs

    @memoized_property
    def props(self):
        return {'id': self.id,
                'value_id': self.value_id,
                'name': self.name,
                'version': self.version,
                'fn_module': self.fn_module,
                'fn_name': self.fn_name,
                'composite': self.composite,
                'value_id_duration': self.value_id_duration,
                'compute_duration': self.compute_duration,
                'hash_duration': self.hash_duration,
                'input_artifact_ids': self.input_artifact_ids,
                'serializer': self.serializer,
                'load_kwargs': self.load_kwargs,
                'dump_kwargs': self.dump_kwargs,
                'custom_fields': self.custom_fields,
                'computed_at': self.computed_at}

    def __repr__(self):
        return '<Artifact %r>' % self.id


class ArtifactSet(Base):
    __tablename__ = 'artifact_sets'

    id = sa.Column(pg.INTEGER, primary_key=True)
    set_id = sa.Column(pg.VARCHAR(SHA1_LENGTH))
    name = sa.Column(pg.VARCHAR(1000))
    created_at = sa.Column(pg.TIMESTAMP, default=datetime.utcnow)


    def __init__(self, artifact_set):
        self.set_id = artifact_set.id
        self.name = artifact_set.name
        self.created_at = artifact_set.created_at

    @memoized_property
    def props(self):
        return {'id': self.set_id,
                'name': self.name,
                'created_at': self.created_at}

    def __repr__(self):
        return '<ArtifactSet %r, %r>' % (self.set_id, self.name)


class ArtifactSetMember(Base):
    __tablename__ = 'artifact_set_members'

    set_id = sa.Column(pg.VARCHAR(SHA1_LENGTH), #sa.ForeignKey("artifact_sets.set_id"),
                       primary_key=True)
    artifact_id = sa.Column(pg.VARCHAR(SHA1_LENGTH), #sa.ForeignKey("artifacts.id"),
                            primary_key=True)



import datetime
import os
import shutil
import time
from collections import namedtuple

import toolz as t
from boltons import funcutils as bfu

from . import artifact_id_hasher as ah
from . import repos as repos
from . import serializers as s
from . import utils
from ._dependencies import dependencies
from .hashing import file_hash, hash
from .serializers import DEFAULT_VALUE_SERIALIZER


class ImpureFunctionError(Exception):
    pass

def get_metadata(f):
    if hasattr(f, '_provenance_metadata'):
        return f._provenance_metadata
    if hasattr(f, 'func'):
        return get_metadata(f.func)
    else:
        return {}


artifact_properties = ['id', 'value_id', 'inputs', 'fn_module', 'fn_name', 'value',
                       'name', 'version', 'composite', 'value_id_duration',
                       'serializer', 'load_kwargs', 'dump_kwargs',
                       'compute_duration', 'hash_duration', 'computed_at',
                       'custom_fields', 'input_artifact_ids', 'run_info']

ArtifactRecord = namedtuple('ArtifactRecord', artifact_properties)



def fn_info(f):
    info = utils.fn_info(f)
    metadata = get_metadata(f)
    info['identifiers'] = {'name': metadata['name'],
                           'version': metadata['version'],
                           'input_hash_fn': metadata['input_hash_fn']}
    info['input_process_fn'] = metadata['input_process_fn']
    info['composite'] = metadata['returns_composite']
    info['archive_file'] = metadata['archive_file']
    info['custom_fields'] = metadata['custom_fields']
    info['preserve_file_ext'] = metadata['preserve_file_ext']

    if info['composite']:
        if info['archive_file']:
            raise NotImplementedError("Using 'composite' and 'archive_file' is not supported.")
        info['serializer'] = metadata['serializer'] or {}
        info['load_kwargs'] = metadata['load_kwargs'] or {}
        info['dump_kwargs'] = metadata['dump_kwargs'] or {}
        valid_serializer = isinstance(info['serializer'], dict)
        for serializer in info['serializer'].values():
            valid_serializer = valid_serializer and serializer in s.serializers
            if not valid_serializer:
                break
    elif info['archive_file']:
        serializer = metadata['serializer'] or 'file'
        if serializer != 'file':
            raise ValueError("With 'archive_file' set True the only valid 'serializer' is 'file'")
        if metadata.get('dump_kwargs') is not None:
            raise ValueError("With 'archive_file' set True you may not specify any dump_kwargs.")
        if metadata.get('load_kwargs') is not None:
            raise ValueError("With 'archive_file' set True you may not specify any load_kwargs.")
        info['serializer'] = 'file'
        info['load_kwargs'] = metadata['load_kwargs'] or {}
        info['dump_kwargs'] = (metadata['dump_kwargs']
                               or {'delete_original': metadata['delete_original_file']})
        info['delete_original_file'] = metadata['delete_original_file']
        valid_serializer = True
    else:
        info['serializer'] = metadata.get('serializer', DEFAULT_VALUE_SERIALIZER.name) or DEFAULT_VALUE_SERIALIZER.name
        info['load_kwargs'] = metadata.get('load_kwargs', None)
        info['dump_kwargs'] = metadata.get('dump_kwargs', None)
        valid_serializer = info['serializer'] in s.serializers

    if not valid_serializer:
        msg = 'Invalid serializer option "{}" for artifact "{}", available serialziers: {} '.\
              format(info['serializer'], info['name'], tuple(s.serializers.keys()))
        raise ValueError(msg)

    return info


def hash_inputs(inputs):
    kargs = {}
    varargs = []
    all_ids = frozenset()

    for k, v in inputs['kargs'].items():
        h, ids = hash(v, hasher=ah.artifact_id_hasher())
        kargs[k] = h
        all_ids |= ids

    for v in inputs['varargs']:
        h, ids = hash(v, hasher=ah.artifact_id_hasher())
        varargs.append(h)
        all_ids |= ids

    input_hashes = {'kargs': kargs, 'varargs': tuple(varargs)}
    return (input_hashes, all_ids)


def create_id(input_hashes, input_hash_fn, name, version):
    return t.thread_first(input_hashes,
                          input_hash_fn,
                          (t.merge, {'name': name, 'version': version}),
                          hash)


@t.curry
def composite_artifact(repo, _run_info, inputs, input_hashes, input_artifact_ids,
                       input_hash_fn, artifact_info, compute_duration,
                       computed_at, key, value):
    start_hash_time = time.time()
    info = artifact_info.copy()
    info['composite'] = False
    info['name'] = '{}_{}'.format(info['name'], key)
    info['serializer'] = info['serializer'].get(key, DEFAULT_VALUE_SERIALIZER.name)
    info['load_kwargs'] = info['load_kwargs'].get(key, None)
    info['dump_kwargs'] = info['dump_kwargs'].get(key, None)

    id = create_id(input_hashes, input_hash_fn, info['name'], info['version'])
    hash_duration = time.time() - start_hash_time

    start_value_id_time = time.time()
    value_id = hash(value)
    value_id_duration = time.time() - start_hash_time

    record = ArtifactRecord(id=id, value_id=value_id, value=value,
                            input_artifact_ids=input_artifact_ids,
                            value_id_duration=value_id_duration,
                            compute_duration=compute_duration,
                            hash_duration=hash_duration, computed_at=computed_at,
                            inputs=inputs, run_info=_run_info, **info)
    return repo.put(record)

def _base_fn(f):
    if utils.is_curry_func(f):
        return utils.inner_function(f)
    else:
        return f


_EXT_MAPPINGS = {'mpeg': 'mpg',
                 'jpeg': 'jpg'}
def _extract_extension(filename):
    ext = os.path.splitext(filename)[1]
    if len(ext) > 0:
        ext = ext.lower().strip()[1:]
        return '.' + _EXT_MAPPINGS.get(ext, ext)
    else:
        return ext


def _archive_file_hash(filename, preserve_file_ext):
    if hasattr(filename, '__fspath__'):
        filename = filename.__fspath__()
    else:
        filename = str(filename)
    if not os.path.exists(filename):
        raise FileNotFoundError("Unable to archive file, {}, because it doesn't exist!".format(filename))
    # TODO: figure out best place to put the hash_name config and use in both cases
    value_id = file_hash(filename)
    if preserve_file_ext:
        extension = _extract_extension(filename)
        value_id += extension
    return value_id


def run_info():
    return repos.Config.current().run_info()


@t.curry
def provenance_wrapper(repo, f):
    base_fn = _base_fn(f)
    extract_args = utils.args_extractor(base_fn, merge_defaults=True)
    func_info = fn_info(f)
    input_process_fn = func_info['input_process_fn']

    artifact_info = {'name': func_info['identifiers']['name'],
                     'version': func_info['identifiers']['version'],
                     'fn_name': func_info['name'], 'fn_module': func_info['module'],
                     'custom_fields': func_info['custom_fields'],
                     'serializer': func_info['serializer'],
                     'load_kwargs': func_info['load_kwargs'],
                     'dump_kwargs': func_info['dump_kwargs'],
                     'composite': func_info['composite']}

    @bfu.wraps(f)
    def _provenance_wrapper(*args, **kargs):
        r = repos.get_default_repo() if repo is None else repo
        _run_info = run_info()
        archive_file = func_info['archive_file']

        start_hash_time = time.time()
        varargs, argsd = extract_args(args, kargs)
        raw_inputs = {'varargs': varargs + func_info['varargs'],
                      'kargs': t.merge(argsd, func_info['kargs'])}
        inputs = input_process_fn(raw_inputs)

        value_id = None
        filename = None
        archive_file_helper = archive_file and '_archive_file_filename' in raw_inputs['kargs']
        if archive_file_helper:
            filename = raw_inputs['kargs']['_archive_file_filename']
            value_id = _archive_file_hash(filename, func_info['preserve_file_ext'])
            inputs['filehash'] = value_id

        input_hashes, input_artifact_ids = hash_inputs(inputs)

        id = create_id(input_hashes, **func_info['identifiers'])
        hash_duration = time.time() - start_hash_time

        try:
            artifact = r.get_by_id(id)
        except KeyError:
            artifact = None

        if artifact is None:
            start_compute_time = time.time()
            computed_at = datetime.datetime.utcnow()
            value = f(*varargs, **argsd)
            compute_duration = time.time() - start_compute_time

            post_input_hashes, _ = hash_inputs(inputs)
            if id != create_id(post_input_hashes, **func_info['identifiers']):
                modified_inputs = []
                kargs = input_hashes['kargs']
                varargs = input_hashes['varargs']
                for name, _hash in post_input_hashes['kargs'].items():
                    if _hash != kargs[name]:
                        modified_inputs.append(name)
                for i, _hash in enumerate(post_input_hashes['varargs']):
                    if _hash != varargs[i]:
                        modified_inputs.append("varargs[{}]".format(i))
                msg = "The {}.{} function modified arguments: ({})".format(
                    func_info['module'], func_info['name'], ",".join(modified_inputs))
                raise ImpureFunctionError(msg)

            if artifact_info['composite']:
                input_hash_fn = func_info['identifiers']['input_hash_fn']
                ca = composite_artifact(r, _run_info, inputs, input_hashes,
                                        input_artifact_ids, input_hash_fn,
                                        artifact_info, compute_duration,
                                        computed_at)
                value = {k: ca(k, v) for k, v in value.items()}
                artifact_info['serializer'] = DEFAULT_VALUE_SERIALIZER.name
                artifact_info['load_kwargs'] = None
                artifact_info['dump_kwargs'] = None


            start_value_id_time = time.time()
            if archive_file:
                if not archive_file_helper:
                    filename = value
                    value_id = _archive_file_hash(filename, func_info['preserve_file_ext'])
                value = ArchivedFile(value_id, filename, in_repo=False)
            else:
                value_id = hash(value)
            value_id_duration = time.time() - start_value_id_time

            record = ArtifactRecord(id=id, value_id=value_id, value=value,
                                    input_artifact_ids=input_artifact_ids,
                                    value_id_duration=value_id_duration,
                                    compute_duration=compute_duration,
                                    hash_duration=hash_duration,
                                    computed_at=computed_at, run_info=_run_info,
                                    inputs=inputs, **artifact_info)
            artifact = r.put(record)

            if archive_file:
                # mark the file as in the repo (yucky, I know)
                artifact.value.in_repo = True

        elif archive_file_helper and func_info.get('delete_original_file', False):
        # if we hit an artifact with archive_file we may still need to clean up the
        # referenced file. This is normally taken care of when the file is 'serialzied'
        # (see file_dump), but in the case of an artifact hit this would never happen.
        # One potential downside of this approach is that this local file will be
        # deleted and if the artifact value (i.e. the existing file) is not local
        # yet it will download the file that we just deleted. Another approach would
        # be to do a a put_overwrite which would potentially upload files multiple times.
        # So for now, the cleanest way is to accept the potential re-downloading of data.
            os.remove(filename)


        return artifact.proxy()

    if utils.is_curry_func(f):
        fb = bfu.FunctionBuilder.from_func(utils.inner_function(f))
        for arg in f.args + tuple(f.keywords.keys()):
            fb.remove_arg(arg)
        param_info = utils.param_info(f)
        args = []
        defaults = []
        for arg, value in param_info.items():
            args.append(arg)
            if value != utils.UNSPECIFIED_ARG:
               defaults.append(value)
        arg_inv = ['{}={}'.format(arg,arg) for arg in args]
        fb.body = 'return _provenance_wrapper(%s)' % ", ".join(arg_inv)
        fb.args = args
        fb.defaults = tuple(defaults)
        execdict = {'_provenance_wrapper': _provenance_wrapper}
        ret = fb.get_func(execdict, with_dict=True)
        return ret
    return _provenance_wrapper


def remove_inputs_fn(to_remove):
    def remove_inputs(inputs):
        kargs = t.keyfilter(lambda k: k not in to_remove, inputs['kargs'])
        return {'kargs': kargs, 'varargs': inputs['varargs']}
    return remove_inputs


def provenance(version=0, repo=None, name=None, merge_defaults=None,
               ignore=None, input_hash_fn=None, remove=None, input_process_fn=None,
               archive_file=False, delete_original_file=False, preserve_file_ext=False,
               returns_composite=False, custom_fields=None,
               serializer=None, load_kwargs=None, dump_kwargs=None,
               _provenance_wrapper=provenance_wrapper):
    if ignore and input_hash_fn:
        raise ValueError("You cannot provide both ignore and input_hash_fn")

    if ignore:
        ignore = frozenset(ignore)
        input_hash_fn = remove_inputs_fn(ignore)

    if not input_hash_fn:
        input_hash_fn = lambda inputs: inputs

    if remove and input_process_fn:
        raise ValueError("You cannot provide both remove and input_process_fn")

    if remove:
        remove = frozenset(remove)
        input_process_fn = remove_inputs_fn(remove)

    if not input_process_fn:
        input_process_fn = lambda inputs: inputs

    def wrapped(f):
        _name = name
        if _name is None:
            _name = f.__name__
        f._provenance_metadata = {'version': version,
                                  'name': _name,
                                  'archive_file': archive_file,
                                  'delete_original_file': delete_original_file,
                                  'input_hash_fn': input_hash_fn,
                                  'input_process_fn': input_process_fn,
                                  'archive_file': archive_file,
                                  'delete_original_file': delete_original_file,
                                  'preserve_file_ext': preserve_file_ext,
                                  'returns_composite': returns_composite,
                                  'archive_file': archive_file,
                                  'custom_fields': custom_fields or {},
                                  'serializer': serializer,
                                  'load_kwargs': load_kwargs,
                                  'dump_kwargs': dump_kwargs}
        f.__merge_defaults__ = merge_defaults
        return _provenance_wrapper(repo, f)

    return wrapped

class ArchivedFile(object):
    def __init__(self, id, original_filename=None, in_repo=True):
        self.blob_id = id
        self.original_filename = original_filename
        self.in_repo = in_repo

    def abspath(self):
        repo = repos.get_default_repo()
        path = repo.blobstore._filename(self.blob_id)
        return os.path.abspath(path)

    def __fspath__(self):
        return self.abspath() if self.in_repo else self.original_filename

    def __str__(self):
        return self.__fspath__()

    def __repr__(self):
        if self.original_filename:
            return "<ArchivedFile {}, {} >".format(self.blob_id, self.original_filename)
        else:
            return "<ArchivedFile {} >".format(self.blob_id)


def file_dump(archived_file, dest_filename, delete_original=False):
    op = shutil.move if delete_original else shutil.copy
    op(archived_file.original_filename, dest_filename)


def file_load(id):
    return ArchivedFile(id, in_repo=True)


s.register_serializer('file', file_dump, file_load)


def archive_file(filename, name=None, delete_original=False, custom_fields=None, preserve_ext=False):
    @provenance(archive_file=True, name=name or 'archive_file', preserve_file_ext=preserve_ext,
                delete_original_file=delete_original, remove=['_archive_file_filename'],
                custom_fields=custom_fields)

    # we want artifacts created by archive_file to be invariant to the
    # filename (see remove above) but # not the custom_fields.
    # filename is still passed in so the hash of the file on disk can be
    # computed as part of the id of the artifact.
    def _archive_file(_archive_file_filename, custom_fields):
        return filename
    return _archive_file(filename, custom_fields)


def provenance_set(set_name=None, initial_set=None, set_name_fn=None):
    if set_name and set_name_fn:
        raise ValueError("You cannot provide both set_name and set_name_fn.")

    def make_wrapper(f):
        if set_name_fn:
            base_fn = _base_fn(f)
            extract_args = utils.args_extractor(base_fn, merge_defaults=True)
            func_info = utils.fn_info(f)

        @bfu.wraps(f)
        def wrapper(*fargs, **fkargs):
            if set_name_fn:
                varargs, argsd = extract_args(fargs, fkargs)
                varargs += func_info['varargs']
                argsd.update(func_info['kargs'])
                name = set_name_fn(*varargs, **argsd)
            else:
                name = set_name

            with repos.capture_set(name=name, initial_set=initial_set) as result:
                f(*fargs, **fkargs)
            return result[0]

        return wrapper

    return make_wrapper


def promote(artifact_or_id, to_repo, from_repo=None):
    from_repo = from_repo if from_repo else repos.get_default_repo()
    artifact = repos.coerce_to_artifact(artifact_or_id, repo=from_repo)
    for a in dependencies(artifact):
        if a not in to_repo:
            to_repo.put(a)



import os
import provenance.core as core
import provenance.blobstores as bs
import provenance.repos as r
import copy
import toolz as t


import logging
logger = logging.getLogger(__name__)


@t.curry
def full_config(configs, base_config):
    if 'type' in base_config:
        return base_config
    prototype = full_config(configs, configs[base_config['prototype']])
    return t.thread_first(prototype,
                          (t.merge, base_config),
                          (t.dissoc, 'prototype'))


def merge_prototypes(config):
    return t.valmap(full_config(config), config)

@t.curry
def atomic_item_from_config(config, type_dict, item_plural, name=None):
    stype = config['type']
    if stype not in type_dict:
        raise Exception("{} may only be created of types: {}, you had {}".
                        format(item_plural, tuple(type_dict.keys()), stype))
    cls = type_dict[stype]
    kargs = t.dissoc(config, 'type')
    return cls(**kargs)


BLOBSTORE_TYPES = {'disk': bs.DiskStore, 's3': bs.S3Store,
                   'memory': bs.MemoryStore, 'chained': bs.ChainedStore}

blobstore_from_config = atomic_item_from_config(type_dict=BLOBSTORE_TYPES,
                                           item_plural='Blobstores')

REPO_TYPES= {'postgres': r.PostgresRepo, 'memory': r.MemoryRepo,
             'chained': r.ChainedRepo}

repo_from_config = atomic_item_from_config(type_dict=REPO_TYPES,
                                           item_plural='Artifact Repos')

def items_from_config(config, atomic_from_config, items_name,
                      item_type, silence_warnings):
    config = merge_prototypes(copy.deepcopy(config))

    atomic_stores = {}
    for k, c in config.items():
        try:
            if c['type'] != 'chained':
                store = atomic_from_config(c, name=k)
                if store:
                    atomic_stores[k] = store
        except Exception as e:
            if not silence_warnings:
                logger.warning("Error creating %s %s from config - Skipping",
                               item_type, k, exc_info=True)

    def create_chained(name, config):
        # resolve the stores
        chained = {n for n in config[items_name] if n in atomic_stores}
        if len(chained) != len(config[items_name]):
            missing_configs = set(config[items_name]) - chained
            if not silence_warnings:
                logger.warning("Skipping chained %s %s due to missing %s: %s",
                               item_type, name, items_name, missing_configs)
            return None

        config[items_name] = [atomic_stores[n] for n in config[items_name]]
        return atomic_from_config(config, name=name)

    chained_stores = {}
    for k, c in config.items():
        try:
            if c['type'] == 'chained':
                store = create_chained(k, c)
                if store:
                    chained_stores[k] = store
        except Exception as e:
            if not silence_warnings:
                logger.warning("Error creating %s %s from config - Skipping",
                               item_type, k, exc_info=True)

    return t.merge(chained_stores, atomic_stores)


def blobstores_from_config(config, silence_warnings=False):
    return items_from_config(config, blobstore_from_config, 'stores', 'blobstore', silence_warnings)


def repos_from_config(config, blobstores, silence_warnings=False):
    def from_config(atomic_config, name):
        if 'store' in atomic_config:
            if not atomic_config['store'] in blobstores:
                if not silence_warnings:
                    logger.warning("Skipping %s repo due to missing store: %s",
                                   name, atomic_config['store'])
                return None

            atomic_config['store'] = blobstores[atomic_config['store']]
        return repo_from_config(atomic_config)
    return items_from_config(config, from_config, 'repos', 'repo', silence_warnings)


def from_config(config):
    silence_warnings = config.get('silence_warnings', False)
    blobstores = blobstores_from_config(config['blobstores'],
                                        silence_warnings)
    repos = repos_from_config(config['artifact_repos'], blobstores,
                              silence_warnings)
    return {'blobstores': blobstores, 'repos': repos}

def load_config(config):
    objs = from_config(config)
    pconfig = r.Config(objs['blobstores'], objs['repos'], config['default_repo'])
    r.Config.set_current(pconfig)
    return pconfig


def load_yaml_config(filename):
    import yaml
    with open(filename, 'r') as f:
        return load_config(yaml.load(f))



"""
Fast cryptographic hash of Python objects, with a special case for fast
hashing of numpy arrays.


This code was originally taken from joblib and modified.

 Author: Gael Varoquaux <gael dot varoquaux at normalesup dot org>
 Copyright (c) 2009 Gael Varoquaux
 License: BSD Style, 3 clauses.
"""

# Author: Gael Varoquaux <gael dot varoquaux at normalesup dot org>
# Copyright (c) 2009 Gael Varoquaux
# License: BSD Style, 3 clauses.

import cloudpickle
from functools import singledispatch
import pickle
import hashlib
import sys
import types
import struct
import io
import decimal

from joblib._compat import _bytes_or_unicode, PY3_OR_LATER



@singledispatch
def value_repr(obj):
    method = getattr(obj, "value_repr", None)
    if callable(method):
        return method()
    else:
        return obj


Pickler = cloudpickle.CloudPickler

class _ConsistentSet(object):
    """ Class used to ensure the hash of Sets is preserved
        whatever the order of its items.
    """
    def __init__(self, _set):
        # Forces order of elements in set to ensure consistent hash.
        self._type = type(_set)
        try:
            # Trying first to order the set assuming the type of elements is
            # consistent and orderable.
            # This fails on python 3 when elements are unorderable
            # but we keep it in a try as it's faster.
            self._sequence = sorted(_set)
        except (TypeError, decimal.InvalidOperation):
            # If elements are unorderable, sorting them using their hash.
            # This is slower but works in any case.
            self._sequence = sorted((hash(e) for e in _set))


class _MyHash(object):
    """ Class used to hash objects that won't normally pickle """

    def __init__(self, *args):
        self.args = args


class Hasher(Pickler):
    """ A subclass of pickler, to do cryptographic hashing, rather than
        pickling.
    """

    def __init__(self, hash_name='sha1'):
        self.stream = io.BytesIO()
        # By default we want a pickle protocol that only changes with
        # the major python version and not the minor one
        protocol = (pickle.DEFAULT_PROTOCOL if PY3_OR_LATER
                    else pickle.HIGHEST_PROTOCOL)
        Pickler.__init__(self, self.stream, protocol=protocol)
        # Initialise the hash obj
        self._hash = hashlib.new(hash_name)

    def hash(self, obj):
        try:
            self.dump(obj)
        except pickle.PicklingError as e:
            e.args += ('PicklingError while hashing %r: %r' % (obj, e),)
            raise
        dumps = self.stream.getvalue()
        self._hash.update(dumps)
        return self._hash.hexdigest()

    def save(self, obj):
        obj = value_repr(obj)
        if isinstance(obj, (types.MethodType, type({}.pop))):
            # the Pickler cannot pickle instance methods; here we decompose
            # them into components that make them uniquely identifiable
            if hasattr(obj, '__func__'):
                func_name = obj.__func__.__name__
            else:
                func_name = obj.__name__
            inst = obj.__self__
            if type(inst) == type(pickle):
                obj = _MyHash(func_name, inst.__name__)
            elif inst is None:
                # type(None) or type(module) do not pickle
                obj = _MyHash(func_name, inst)
            else:
                cls = obj.__self__.__class__
                obj = _MyHash(func_name, inst, cls)
        Pickler.save(self, obj)

    def memoize(self, obj):
        # don't memoize so that the hashes are completely value-based
        return

    # The dispatch table of the pickler is not accessible in Python
    # 3, as these lines are only bugware for IPython, we skip them.
    def save_global(self, obj, name=None, pack=struct.pack):
        # We have to override this method in order to deal with objects
        # defined interactively in IPython that are not injected in
        # __main__
        kwargs = dict(name=name, pack=pack)
        if sys.version_info >= (3, 4):
            del kwargs['pack']
        try:
            Pickler.save_global(self, obj, **kwargs)
        except pickle.PicklingError:
            Pickler.save_global(self, obj, **kwargs)
            module = getattr(obj, "__module__", None)
            if module == '__main__':
                my_name = name
                if my_name is None:
                    my_name = obj.__name__
                mod = sys.modules[module]
                if not hasattr(mod, my_name):
                    # IPython doesn't inject the variables define
                    # interactively in __main__
                    setattr(mod, my_name, obj)

    dispatch = Pickler.dispatch.copy()
    # builtin
    dispatch[type(len)] = save_global
    # type
    dispatch[type(object)] = save_global
    # classobj
    dispatch[type(Pickler)] = save_global
    # function
    dispatch[type(pickle.dump)] = save_global

    def _batch_setitems(self, items):
        # forces order of keys in dict to ensure consistent hash.
        try:
            # Trying first to compare dict assuming the type of keys is
            # consistent and orderable.
            # This fails on python 3 when keys are unorderable
            # but we keep it in a try as it's faster.
            Pickler._batch_setitems(self, iter(sorted(items)))
        except TypeError:
            # If keys are unorderable, sorting them using their hash. This is
            # slower but works in any case.
            Pickler._batch_setitems(self, iter(sorted((hash(k), v)
                                                      for k, v in items)))

    def save_set(self, set_items):
        # forces order of items in Set to ensure consistent hash
        Pickler.save(self, _ConsistentSet(set_items))

    dispatch[type(set())] = save_set
    dispatch[type(frozenset())] = save_set


class NumpyHasher(Hasher):
    """ Special case the hasher for when numpy is loaded.
    """

    def __init__(self, hash_name='sha1', coerce_mmap=True):
        """
            Parameters
            ----------
            hash_name: string
                The hash algorithm to be used
            coerce_mmap: boolean
                Make no difference between np.memmap and np.ndarray
                objects.
        """
        self.coerce_mmap = coerce_mmap
        self.chunk_size = 200 * 1024 * 1024 # 200 Mb
        Hasher.__init__(self, hash_name=hash_name)
        # delayed import of numpy, to avoid tight coupling
        import numpy as np
        self.np = np

    def hash_array(self, a):
        self._hash.update(a.tobytes())

    def save(self, obj):
        """ Subclass the save method, to hash ndarray subclass, rather
            than pickling them. Off course, this is a total abuse of
            the Pickler class.
        """
        if isinstance(obj, self.np.ndarray) and not obj.dtype.hasobject:
            # Compute a hash of the object
            obj_bytes = obj.dtype.itemsize * obj.size
            if obj_bytes > self.chunk_size:
                # For arrays larger than `self.chunk_size` we will attempt
                # to change the shape of a shallow copy and then hash the data
                # in chunks
                try:
                    copy = obj[:]
                    copy.shape = (copy.size,)
                except AttributeError as e:
                    if e.args[0] != 'incompatible shape for a non-contiguous array':
                        raise e

                    # TODO: I am punting here for now and do a reshape that will make
                    # a copy, but it could be possible to get the bytes out of obj
                    # without needing one
                    copy = obj.reshape((obj.size,))

                i = 0; size = copy.size
                typed_chunk_size = self.chunk_size // copy.dtype.itemsize
                while i < size:
                    end = min(i + typed_chunk_size, size)
                    self.hash_array(copy[i:end])
                    i = end

            else:
                # Small arrays are hashed all at once
                self.hash_array(obj)

            # We store the class, to be able to distinguish between
            # Objects with the same binary content, but different
            # classes.
            if self.coerce_mmap and isinstance(obj, self.np.memmap):
                # We don't make the difference between memmap and
                # normal ndarrays, to be able to reload previously
                # computed results with memmap.
                klass = self.np.ndarray
            else:
                klass = obj.__class__
            # We also return the dtype and the shape, to distinguish
            # different views on the same data with different dtypes.

            # The object will be pickled by the pickler hashed at the end.
            obj = (klass, ('HASHED', obj.dtype, obj.shape))
        elif isinstance(obj, self.np.dtype):
            # Atomic dtype objects are interned by their default constructor:
            # np.dtype('f8') is np.dtype('f8')
            # This interning is not maintained by a
            # pickle.loads + pickle.dumps cycle, because __reduce__
            # uses copy=True in the dtype constructor. This
            # non-deterministic behavior causes the internal memoizer
            # of the hasher to generate different hash values
            # depending on the history of the dtype object.
            # To prevent the hash from being sensitive to this, we use
            # .descr which is a full (and never interned) description of
            # the array dtype according to the numpy doc.
            klass = obj.__class__
            obj = (klass, ('HASHED', obj.descr))
        Hasher.save(self, obj)


def hash(obj, hasher=None, hash_name='sha1', coerce_mmap=True):
    """ Quick calculation of a hash to identify uniquely Python objects
        containing numpy arrays. The difference with this hash and joblib
        is that it tries to hash different mutable objects with the same
        values to the same hash.


        Parameters
        -----------
        hash_name: 'md5' or 'sha1'
            Hashing algorithm used. sha1 is supposedly safer, but md5 is
            faster.
        coerce_mmap: boolean
            Make no difference between np.memmap and np.ndarray
    """
    if hasher is None:
        if 'numpy' in sys.modules:
            hasher = NumpyHasher(hash_name=hash_name, coerce_mmap=coerce_mmap)
        else:
            hasher = Hasher(hash_name=hash_name)

    return hasher.hash(obj)


def file_hash(filename, hash_name='sha1'):
    """Streams the bytes of the given file through either md5 or sha1
       and returns the hexdigest.
    """
    if hash_name not in set(['md5', 'sha1']):
        raise ValueError('hashname must be "md5" or "sha1"')

    hasher = hashlib.md5() if hash_name == 'md5' else hashlib.sha1()
    with open(filename, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hasher.update(chunk)
    return hasher.hexdigest()



import cloudpickle
import io
from joblib._compat import _bytes_or_unicode, PY3_OR_LATER
from ordered_set import OrderedSet
import pickle

from . import repos as r
from .compatibility import string_type


Pickler = cloudpickle.CloudPickler


class DependencyWalker(Pickler):
    def __init__(self):
        self.stream = io.BytesIO()
        self.dependents = []
        self.branches = []
        protocol = (pickle.DEFAULT_PROTOCOL if PY3_OR_LATER
                    else pickle.HIGHEST_PROTOCOL)
        Pickler.__init__(self, self.stream, protocol=protocol)

    def save(self, obj):
        if isinstance(obj, r.Artifact):
            self.dependents.append(obj)
        elif r.is_proxy(obj):
            self.dependents.append(obj.artifact)
        else:
            Pickler.save(self, obj)

    def deps(self, artifact):
        self.dependents = []
        self.dump(artifact)
        return self.dependents


def _deps(val):
    return DependencyWalker().deps(val)


def _artifact_branches(artifact):
    if artifact.composite:
        objs = _deps(artifact.inputs) + _deps(artifact.value)
    else:
        objs = _deps(artifact.inputs)
    objs.sort(key=lambda a: a.id)
    return objs


def dependencies(artifact_or_id):
    """
    Returns a reversed breadth first search. This guarantees that
    for all artifacts in the list. All of an artifacts dependencies
    will come before it.
    """
    artifact = r.coerce_to_artifact(artifact_or_id)
    visited = []
    queue = [artifact]
    while queue:
        a, *queue = queue

        if a in visited:
            continue

        visited.append(a)
        queue.extend(_artifact_branches(a))

    visited.reverse()
    return visited



import sys

from . import hashing as h
from . import repos as r


def _save(obj, ids):
    if isinstance(obj, r.Artifact):
        ids.add(obj.id)
    if r.is_proxy(obj):
        ids.add(obj.artifact.id)


class ArtifactIdHasher(h.Hasher):
    def __init__(self, ids=None, hash_name='sha1'):
        if ids is None:
            ids = set()

        self.ids = ids
        h.Hasher.__init__(self, hash_name=hash_name)

    def save(self, obj):
        _save(obj, self.ids)
        h.Hasher.save(self, obj)

    def hash(self, obj):
        return (h.Hasher.hash(self, obj), frozenset(self.ids))


class NumpyArtifactIdHasher(h.NumpyHasher):
    def __init__(self, ids=None, hash_name='sha1', coerce_mmap=True):
        if ids is None:
            ids = set()

        self.ids = ids
        h.NumpyHasher.__init__(self, hash_name=hash_name, coerce_mmap=coerce_mmap)

    def save(self, obj):
        _save(obj, self.ids)
        h.NumpyHasher.save(self, obj)

    def hash(self, obj):
        return (h.NumpyHasher.hash(self, obj), frozenset(self.ids))


def artifact_id_hasher(*args, **kwargs):
    if 'numpy' in sys.modules:
        return NumpyArtifactIdHasher(*args, **kwargs)
    else:
        return ArtifactIdHasher(*args, **kwargs)



from collections import namedtuple
import toolz as t
import cloudpickle
import joblib

from .hashing import hash

def cloudpickle_dump(obj, filename, **kargs):
    with open(filename, 'wb') as f:
        return cloudpickle.dump(obj, f, **kargs)


def cloudpickle_load(filename, **kargs):
    with open(filename, 'rb') as f:
        return cloudpickle.load(f, **kargs)


Serializer = namedtuple('Serializer', 'name, dump, load')


serializers = {'joblib': Serializer('joblib', joblib.dump, joblib.load),
               'cloudpickle': Serializer('cloudpickle',
                                         cloudpickle_dump,
                                         cloudpickle_load),}


def register_serializer(name, dump, load):
    serializers[name] = Serializer(name, dump, load)

@t.memoize(key=lambda *args: hash(args))
def partial_serializer(serializer_name, dump_kwargs, load_kwargs):
    s = serializers[serializer_name]
    return Serializer(s.name,
                      t.partial(s.dump, **dump_kwargs) if dump_kwargs else s.dump,
                      t.partial(s.load, **load_kwargs) if load_kwargs else s.load)


def serializer(artifact):
    return partial_serializer(artifact.serializer,
                              artifact.dump_kwargs,
                              artifact.load_kwargs)

DEFAULT_VALUE_SERIALIZER = serializers['joblib']
DEFAULT_INPUT_SERIALIZER = serializers['joblib']



#!/usr/bin/env python

from os.path import exists
from setuptools import setup

import versioneer

setup(
    name='provenance',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    packages=('provenance',),
    setup_requires=['pytest>=3.0.0', 'pytest-runner'],
    install_requires=[open('requirements.txt').read().strip().split('\n')],
    tests_require=[open('test_requirements.txt').read().strip().split('\n')],
    description="Provenance and caching library for functions",
    long_description=(open('README.rst').read() if exists('README.rst')
                      else ''),
    author="Ben Mabey",
    author_email="ben@benmabey.com",
    url="http://github.com/Savvysherpa/provenance",
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        "License :: OSI Approved :: MIT License",
    ],
)



import toolz as t
import operator as op


class PermissionError(Exception):
    def __init__(self, action, store, permission):
        message = "A `{}` operation was attempted on {} and {} is set to `False`!".\
                  format(action, store, permission)
        self.action = action
        self.store = store
        self.permission = permission
        Exception.__init__(self, message)


class KeyExistsError(Exception):
    def __init__(self, key, store):
        msg = "The key {} is already present in {}, you can not overwrite it!".\
              format(key, store)
        self.key = key
        self.store = store
        Exception.__init__(self, msg)


class InconsistentKeyError(Exception):
    def __init__(self, key, store, value):
        msg = "The key {} already represents a different value in {}".format(key, store)
        self.key = key
        self.store = store
        self.value = value
        Exception.__init__(self, msg)


def ensure_read(obj, action='get'):
    if not obj._read:
        raise PermissionError(action, obj, 'read')


def ensure_write(obj, action='put'):
    if not obj._write:
        raise PermissionError(action, obj, 'write')

ensure_contains = t.partial(ensure_read, action='contains')

def ensure_present(obj, id):
    if id not in obj:
        raise KeyError(id, obj)

def ensure_delete(obj, id=None, check_contains=True):
    if not obj._delete:
        raise PermissionError('delete', obj, 'delete')
    if check_contains and id is not None and id not in obj:
        raise KeyError(id, obj)


def ensure_put(obj, id, read_through=None, check_contains=True):
    if read_through:
        if not obj._read_through_write:
            raise PermissionError('read_through_put', obj, 'read_through_write')
    elif not obj._write:
        raise PermissionError('put', obj, 'write')
    if check_contains and id in obj:
        raise KeyExistsError(id, obj)

def chained_contains(chained, id, contains=op.contains):
    stores_with_read = [s for s in chained.stores if s._read]
    if len(stores_with_read) == 0:
        raise PermissionError('contains', chained, 'read')

    for store in stores_with_read:
        if store._read and contains(store, id):
            return True
    return False

def chained_put(chained, id, value, put=None, contains=op.contains, **kargs):
    stores_with_write = [s for s in chained.stores if s._write]
    if len(stores_with_write) == 0:
        raise PermissionError('put', chained, 'write')

    record = None
    putin = []
    for store in stores_with_write:
        if not contains(store, id):
            if put:
                record = put(store, id, value, **kargs)
            else:
                record = store.put(id, value, **kargs)
            putin.append(store)

    if (len(putin) == 0 and
        len(stores_with_write) > 0):
        raise KeyExistsError(id, chained)

    return record


def chained_get(chained, get, id, put=None):
    stores_with_read = [s for s in chained.stores if s._read]
    if len(stores_with_read) == 0:
        raise KeyError(id, chained)

    pushback = []
    for store in stores_with_read:
        try:
            value = get(store, id)
            break
        except KeyError:
            if store._read_through_write:
                pushback.append(store)
    else:
        raise KeyError(id, chained)

    for store in pushback:
        if put:
            put(store, id, value, read_through=True)
        else:
            store.put(id, value, read_through=True)
    return value


def chained_delete(chained, id, delete=None, contains=op.contains):
    stores_with_delete = [s for s in chained.stores if s._delete]
    if len(stores_with_delete) == 0:
        raise PermissionError('delete', chained, 'delete')

    foundin = []
    for store in stores_with_delete:
        if contains(store, id):
            foundin.append(store)
            if delete:
                delete(store, id)
            else:
                store.delete(id)
    if len(foundin) == 0:
        raise KeyError(id, chained)
    else:
        return foundin



import copy
import json
import multiprocessing
import operator as ops
import os
import platform
from collections import namedtuple
from contextlib import contextmanager
from datetime import datetime

import numpy as np
import psutil
import sqlalchemy
import sqlalchemy.dialects.postgresql as pg
import sqlalchemy.orm
import toolz as t
import wrapt
from memoized_property import memoized_property

from . import _commonstore as cs
from . import models as db
from . import serializers as s
from . import utils
from .compatibility import string_type
from .hashing import hash, value_repr


def _host_info():
    return {'machine': platform.machine(),
            'nodename': platform.node(),
            'platform': platform.platform(),
            'processor': platform.processor(),
            'cpu_count': multiprocessing.cpu_count(),
            'release': platform.release(),
            'system': platform.system(),
            'version': platform.version()}


def _process_info():
    pid = os.getpid()
    p = psutil.Process(pid)
    return {'cmdline': p.cmdline(),
            'cwd': p.cwd(),
            'exe': p.exe(),
            'name': p.name(),
            'num_fds': p.num_fds(),
            'num_threads': p.num_threads()}

class Config(object):

    _current = None

    @classmethod
    def current(cls):
        return cls._current

    @classmethod
    def set_current(cls, registry):
        cls._current = registry

    def __init__(self, blobstores, repos, default_repo, run_info_fn=None):
        self.blobstores = blobstores
        self.repos = repos
        self.set_default_repo(default_repo)
        self._run_info = None
        self.run_info_fn = run_info_fn or t.identity

    def set_default_repo(self, repo):
        if isinstance(repo, string_type):
            if repo not in self.repos:
                raise Exception("There is no registered repo named '{}'.".format(repo))
            self.default_repo = self.repos[repo]
        else:
            self.default_repo = repo

    def set_run_info_fn(self, fn):
        self.run_info_fn = fn or t.identity
        self._run_info = None

    def run_info(self):
        if self._run_info is None:
            run_info = self.run_info_fn(
                {'host': _host_info(), 'process': _process_info(),
                 'created_at': datetime.utcnow()})
            run_info['id'] = hash(run_info)
            self._run_info = run_info
        return self._run_info


Config.set_current(Config({}, {}, None))


def set_default_repo(repo_or_name):
    Config.current().set_default_repo(repo_or_name)


def get_default_repo():
    return Config.current().default_repo


def set_run_info_fn(fn):
    Config.current().set_run_info_fn(fn)


@contextmanager
def using_repo(repo_or_name):
    prev_repo = get_default_repo()
    set_default_repo(repo_or_name)
    try:
        yield
    finally:
        set_default_repo(prev_repo)


def load_artifact(artifact_id):
    return get_default_repo().get_by_id(artifact_id)


def load_proxy(artifact_id):
    return get_default_repo().get_by_id(artifact_id).proxy()


def get_set_by_id(set_id):
    return get_default_repo().get_set_by_id(set_id)


def get_set_by_name(set_name):
    return get_default_repo().get_set_by_name(set_name)


def create_set(artifact_ids, name=None):
    return ArtifactSet(artifact_ids, name).put()


def name_set(artifact_set_or_id, name):
    repo = get_default_repo()
    if isinstance(artifact_set_or_id, ArtifactSet):
        artifact_set = artifact_set_or_id
    else:
        artifact_set = repo.get_set_by_id(artifact_set_or_id)

    return artifact_set.rename(name).put(repo)


def transform_value(proxy_artifact, transformer_fn):
    """
    Transforms the underlying value of the `proxy_artifact` with
    the provided `transformer_fn`. A new ArtifactProxy is returned
    with the transformed value but with the original artifact.

    The motivation behind this function is to allow archived files
    to be loaded into memory and passed around while preserving
    the provenance of the artifact. It could be used in any other
    situation where you want a different representation of an
    artifact value while allowing the provenance to be tracked.

    Care should be taken when using this function however because
    it will prevent you from reproducing exact artifacts from
    the lineage since this transformer_fn will not be tracked.
    """
    transformed = copy.copy(proxy_artifact)
    transformed.__wrapped__ = transformer_fn(transformed)
    return transformed


class Proxy():
    def value_repr(self):
        return value_repr(self.artifact.value)

    def transform_value(self, transformer_fn):
        return transform_value(self, transformer_fn)


class ArtifactProxy(wrapt.ObjectProxy, Proxy):
    def __init__(self, value, artifact):
        super(ArtifactProxy, self).__init__(value)
        self._self_artifact = artifact

    @property
    def artifact(self):
        return self._self_artifact

    def __repr__(self):
        return '<provenance.ArtifactProxy({}) {} >'.\
            format(self.artifact.id, repr(self.__wrapped__))

    def __reduce__(self):
        return (load_proxy , (self.artifact.id,))


class CallableArtifactProxy(wrapt.CallableObjectProxy, Proxy):
    def __init__(self, value, artifact):
        super(CallableArtifactProxy, self).__init__(value)
        self._self_artifact = artifact

    @property
    def artifact(self):
        return self._self_artifact

    def __repr__(self):
        return '<provenance.ArtifactProxy({}) {} >'.\
            format(self.artifact.id, repr(self.__wrapped__))

    def __reduce__(self):
        return (load_proxy, (self.artifact.id,))


def artifact_proxy(value, artifact):
    if callable(value):
        return CallableArtifactProxy(value, artifact)
    return ArtifactProxy(value, artifact)


def is_proxy(obj):
    return (type(obj) == ArtifactProxy or
            type(obj) == CallableArtifactProxy)



class Artifact(object):
    def __init__(self, repo, props, value=None, inputs=None, run_info=None):
        assert ('id' in props), "props must contain 'id'"
        assert ('value_id' in props), "props must contain 'value_id'"
        self.__dict__ = props.copy()
        self.repo = repo

        if value is not None:
            self._value = value
        if inputs is not None:
            self._inputs = inputs
        if run_info is not None:
            self._run_info = run_info

    @memoized_property
    def value(self):
        return self.repo.get_value(self)

    @memoized_property
    def inputs(self):
        return self.repo.get_inputs(self)

    @memoized_property
    def run_info(self):
        return self.repo.run_info(self.id)

    def proxy(self):
        if self.composite:
            value = lazy_dict(t.valmap(lambda a: lambda: a.proxy(), self.value))
        else:
            value = self.value
        return artifact_proxy(value, self)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.id == other.id
        return NotImplemented

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(('--artifact--', self.id))

    def __repr__(self):
        return '<provenance.Artifact({})>'.format(self.id)

    def __reduce__(self):
        return (load_artifact, (self.id,))


@value_repr.register(Artifact)
def _(artifact):
    return ('artifact', artifact.id)


def _artifact_id(artifact_or_id):
    if isinstance(artifact_or_id, string_type):
        return artifact_or_id
    if hasattr(artifact_or_id, 'id'):
        return artifact_or_id.id
    if hasattr(artifact_or_id, 'artifact'):
        return artifact_or_id.artifact.id
    raise Exception("Unable to coerce into an artifact id: {}".\
                    format(artifact_or_id))


def _artifact_from_record(repo, record):
    if isinstance(record, Artifact):
        return record
    return Artifact(repo,
                    t.dissoc(record._asdict(), 'value', 'inputs', 'run_info'),
                    value=record.value, inputs=record.inputs,
                    run_info=record.run_info)



class ArtifactRepository(object):
    def __init__(self, read=True, write=True, read_through_write=True,
                 delete=False):
        self._read = read
        self._write = write
        self._read_through_write = read_through_write
        self._delete = delete

    def __getitem__(self, artifact_id):
        return self.get_by_id(artifact_id)

    def batch_get_by_id(self, artifact_ids):
        cs.ensure_read(self)
        return [self.get_by_id(id) for id in artifact_ids]


def _find_first(pred, seq):
    for i in seq:
        if pred(i):
            return i


class MemoryRepo(ArtifactRepository):
    def __init__(self, artifacts=None,
                 read=True, write=True, read_through_write=True, delete=True):
        super(MemoryRepo, self).__init__(read=read, write=write,
                                         read_through_write=read_through_write,
                                         delete=delete)
        self.artifacts = artifacts if artifacts else []
        self.sets = []

    def __contains__(self, artifact_or_id):
        cs.ensure_contains(self)
        artifact_id = _artifact_id(artifact_or_id)
        if _find_first(lambda a: a.id == artifact_id, self.artifacts):
            return True
        else:
            return False

    def put(self, record, read_through=False):
        artifact_id = _artifact_id(record)
        cs.ensure_put(self, artifact_id, read_through)
        self.artifacts.append(record)
        return _artifact_from_record(self, record)

    def get_by_id(self, artifact_id):
        cs.ensure_read(self)
        record = _find_first(lambda a: a.id == artifact_id, self.artifacts)
        if record:
            return _artifact_from_record(self, record)
        else:
            raise KeyError(artifact_id, self)

    def get_by_value_id(self, value_id):
        cs.ensure_read(self)
        record = _find_first(lambda a: a.value_id == value_id, self.artifacts)
        if record:
            return _artifact_from_record(self, record)
        else:
            raise KeyError(value_id, self)

    def get_value(self, artifact, composite=False):
        cs.ensure_read(self)
        return _find_first(lambda a: a.id == artifact.id, self.artifacts).value

    def get_inputs(self, artifact):
        cs.ensure_read(self)
        return _find_first(lambda a: a.id == artifact.id, self.artifacts).inputs

    def delete(self, artifact_or_id):
        artifact_id = _artifact_id(artifact_or_id)
        cs.ensure_delete(self)
        new_artifacts = list(t.filter(lambda a: a.id != artifact_id,
                                 self.artifacts))
        if len(new_artifacts) == len(self.artifacts):
            raise KeyError(artifact_id, self)
        else:
            self.artifacts = new_artifacts

    def contains_set(self, set_id):
        art_set = _find_first(lambda s: s.id == set_id, self.sets)
        return True if art_set else False

    def get_set_by_id(self, set_id):
        cs.ensure_read(self)
        art_set = _find_first(lambda s: s.id == set_id, self.sets)
        if not art_set:
            raise KeyError(self, set_id)

        return art_set

    def get_set_by_name(self, name):
        cs.ensure_read(self)
        versions = [s for s in self.sets if s.name == name]
        if not versions:
            raise KeyError(name, self)
        return sorted(versions, key=lambda s: s.created_at,
                      reverse=True)[0]

    def put_set(self, artifact_set, read_through=False):
        cs.ensure_write(self, 'put_set')
        self.sets.append(artifact_set)
        return artifact_set

    def delete_set(self, set_id):
        cs.ensure_delete(self, check_contains=False)
        prev_count = len(self.sets)
        self.sets = [s for s in self.sets if s.id != set_id]
        if len(self.sets) == prev_count:
            raise KeyError(set_id, self)


def _transform(val):
    if isinstance(val, (Artifact)):
        return {'id': val.id, 'type': 'Artifact', 'name': val.name}
    elif type(val) in {ArtifactProxy, CallableArtifactProxy}:
        return {'id': val.artifact.id, 'type': 'ArtifactProxy', 'name': val.artifact.name}
    else:
        return val


def _inputs_json(inputs):
    expanded = t.valmap(_transform, inputs['kargs'])
    expanded['__varargs'] = list(t.map(_transform, inputs['varargs']))

    return expanded


def _ping_postgres(conn, branch):
    """
    Code taken from example here: http://docs.sqlalchemy.org/en/latest/core/pooling.html#dealing-with-disconnects
    """
    if branch:
        # "branch" refers to a sub-connection of a connection,
        # we don't want to bother pinging on these.
        return

    # turn off "close with result".  This flag is only used with
    # "connectionless" execution, otherwise will be False in any case
    save_should_close_with_result = conn.should_close_with_result
    conn.should_close_with_result = False

    try:
        # run a SELECT 1.   use a core select() so that
        # the SELECT of a scalar value without a table is
        # appropriately formatted for the backend
        conn.scalar(sqlalchemy.select([1]))
    except sqlalchemy.exc.DBAPIError as err:
        # catch SQLAlchemy's DBAPIError, which is a wrapper
        # for the DBAPI's exception.  It includes a .connection_invalidated
        # attribute which specifies if this connection is a "disconnect"
        # condition, which is based on inspection of the original exception
        # by the dialect in use.
        if err.connection_invalidated:
            # run the same SELECT again - the connection will re-validate
            # itself and establish a new connection.  The disconnect detection
            # here also causes the whole connection pool to be invalidated
            # so that all stale connections are discarded.
            conn.scalar(sqlalchemy.select([1]))
        else:
            raise
    finally:
        # restore "close with result"
        conn.should_close_with_result = save_should_close_with_result


def _record_pid(dbapi_connection, connection_record):
    connection_record.info['pid'] = os.getpid()


def _check_pid(dbapi_connection, connection_record, connection_proxy):
    pid = os.getpid()
    if connection_record.info['pid'] != pid:
        connection_record.connection = connection_proxy.connection = None
        raise sqlalchemy.exc.DisconnectionError(
            "Connection record belongs to pid %s, "
            "attempting to check out in pid %s" %
            (connection_record.info['pid'], pid))


def _insert_set_members_sql(artifact_set):
    pairs = [(artifact_set.id, id) for id in artifact_set.artifact_ids]
    return """
INSERT INTO artifact_set_members (set_id, artifact_id)
VALUES
{}
ON CONFLICT DO NOTHING
    """.strip().format(",\n".join(t.map(str,pairs)))


class Encoder(json.JSONEncoder):
    def default(self, val):
        if isinstance(val, (datetime)):
            return str(val)
        elif isinstance(val, np.integer):
            return int(val)
        elif isinstance(val, np.floating):
            return float(val)
        elif isinstance(val, np.bool_):
            return bool(val)
        elif isinstance(val, np.ndarray):
            return val.tolist()
        elif callable(val):
            try:
                return utils.fn_info(val)
            except:
                pass
        else:
            try:
                return super(Encoder, self).default(val)
            except Exception as e:
                print("Could not serialize type: {}".format(type(val)))
                return str(type(val))


class PostgresRepo(ArtifactRepository):
    def __init__(self, db, store,
                 read=True, write=True, read_through_write=True, delete=True):
        super(PostgresRepo, self).__init__(read=read, write=write,
                                           read_through_write=read_through_write,
                                           delete=delete)
        self._run = None
        if isinstance(db, string_type):
            self._sessionmaker = self._create_sessionmaker(db)
        else:
            self._session = db

        self.blobstore = store

    @contextmanager
    def session(self):
        if hasattr(self, '_session'):
            close = False
        else:
            self._session = self._sessionmaker()
            close = True

        try:
            yield self._session
        except:
            self._session.rollback()
            raise
        finally:
            if close:
                self._session.close()
                del self._session

    def _create_sessionmaker(self, conn_string):
        db_engine = sqlalchemy.create_engine(conn_string, json_serializer=Encoder().encode)
        sqlalchemy.event.listens_for(db_engine, "engine_connect")(_ping_postgres)
        sqlalchemy.event.listens_for(db_engine, "connect")(_record_pid)
        sqlalchemy.event.listens_for(db_engine, "checkout")(_check_pid)
        return sqlalchemy.orm.sessionmaker(bind=db_engine)

    def __contains__(self, artifact_or_id):
        cs.ensure_contains(self)
        artifact_id = _artifact_id(artifact_or_id)
        with self.session() as s:
            return s.query(db.Artifact).filter(db.Artifact.id == artifact_id).count() > 0

    def _upsert_run(self, session, info):
        sql = pg.insert(db.Run).values(
            id=info['id'], info=info,
            hostname=info['host']['nodename'],
            created_at=info['created_at']
        ).on_conflict_do_nothing(index_elements=['id'])

        session.execute(sql)

        return db.Run(info)

    def put(self, artifact_record, read_through=False):
        with self.session() as session:
            cs.ensure_put(self, artifact_record.id, read_through)
            self.blobstore.put(artifact_record.id, artifact_record.inputs,
                               s.DEFAULT_INPUT_SERIALIZER)
            self.blobstore.put(artifact_record.value_id, artifact_record.value,
                               s.serializer(artifact_record))

            inputs_json = _inputs_json(artifact_record.inputs)
            run = self._upsert_run(session, artifact_record.run_info)
            db_artifact = db.Artifact(artifact_record, inputs_json, run)

            session.add(db_artifact)
            session.commit()

            return _artifact_from_record(self, artifact_record)

    def get_by_id(self, artifact_id):
        cs.ensure_read(self)
        with self.session() as session:
            result = session.query(db.Artifact).filter(db.Artifact.id == artifact_id).first()

        if result:
            return Artifact(self, result.props)
        else:
            raise KeyError(artifact_id, self)

    def batch_get_by_id(self, artifact_ids):
        cs.ensure_read(self)
        with self.session() as session:
            results = session.query(db.Artifact).filter(db.Artifact.id.in_(artifact_ids)).all()

        if len(results) == len(artifact_ids):
            return [Artifact(self, result.props) for result in results]
        else:
            ids = set(artifact_ids)
            found = set([a.id for a in results])
            missing = ids - found
            raise KeyError(missing, self)

    def get_by_value_id(self, value_id):
        cs.ensure_read(self)
        with self.session() as session:
            result = session.query(db.Artifact).filter(db.Artifact.value_id == value_id).first()

        if result:
            return Artifact(self, result.props)
        else:
            raise KeyError(value_id, self)

    def get_value(self, artifact):
        cs.ensure_read(self)
        return self.blobstore.get(artifact.value_id, s.serializer(artifact))

    def get_inputs(self, artifact):
        cs.ensure_read(self)
        return self.blobstore.get(artifact.id, s.DEFAULT_INPUT_SERIALIZER)

    def delete(self, artifact_or_id):
        with self.session() as session:
            cs.ensure_delete(self)
            artifact = self.get_by_id(artifact_or_id)
            (session.query(db.Artifact).
             filter(db.Artifact.id == artifact.id).delete())
            self.blobstore.delete(artifact.id)
            self.blobstore.delete(artifact.value_id)
            session.commit()

    def put_set(self, artifact_set, read_through=False):
        with self.session() as session:
            cs.ensure_write(self, 'put_set')
            db_set = db.ArtifactSet(artifact_set)
            session.add(db_set)
            session.execute(_insert_set_members_sql(artifact_set))
            session.commit()

            return artifact_set

    def _db_to_mem_set(self, result):
        with self.session() as session:
            members = (session.query(db.ArtifactSetMember)
                       .filter(db.ArtifactSetMember.set_id == result.set_id)
                       .all())
            props = result.props
            props['artifact_ids'] = [m.artifact_id for m in members]
            return ArtifactSet(**props)

    def contains_set(self, set_id):
        with self.session() as session:
            if (session.query(db.ArtifactSet)
                      .filter(db.ArtifactSet.set_id == set_id).count() > 0):
                return True
            else:
                return False

    def get_set_by_id(self, set_id):
        cs.ensure_read(self)
        with self.session() as session:
            result = (session.query(db.ArtifactSet)
                      .filter(db.ArtifactSet.set_id == set_id).first())

        if result:
            return self._db_to_mem_set(result)
        else:
            raise KeyError(set_id, self)


    def get_set_by_name(self, name):
        cs.ensure_read(self)
        with self.session() as session:
            result = (session.query(db.ArtifactSet)
                      .filter(db.ArtifactSet.name == name)
                      .order_by(db.ArtifactSet.created_at.desc())
                      .first())

        if result:
            return self._db_to_mem_set(result)
        else:
            raise KeyError(name, self)

    def delete_set(self, set_id):
        cs.ensure_delete(self, check_contains=False)
        with self.session() as session:
            num_deleted = (session.query(db.ArtifactSet).
             filter(db.ArtifactSet.set_id == set_id).delete())
            (session.query(db.ArtifactSetMember).
             filter(db.ArtifactSetMember.set_id == set_id).delete())

        if num_deleted == 0:
            raise KeyError(set_id, self)

    def run_info(self, artifact_id):
        with self.session() as session:
            result = (session.query(db.Run)
                      .filter(db.Run.id == db.Artifact.run_id)
                      .filter(db.Artifact.id == artifact_id)
                      .first())
            return result.info_with_datetimes


DbRepo = PostgresRepo


def _put_only_value(store, id, value, **kargs):
    return store.put(value, **kargs)


def _put_set(store, id, value, **kargs):
    return store.put_set(value, **kargs)


def _contains_set(store, id):
    return store.contains_set(id)


def _delete_set(store, id):
    return store.delete_set(id)


class ChainedRepo(ArtifactRepository):
    def __init__(self, repos):
        self.stores = repos

    def __contains__(self, id):
        return cs.chained_contains(self, id)

    def put(self, record):
        return cs.chained_put(self, record.id, record, put=_put_only_value)

    def put_set(self, artifact_set, read_through=False):
        return cs.chained_put(self, None, artifact_set,
                              contains=_contains_set, put=_put_set)

    def get_by_id(self, artifact_id):
        def get(store, id):
            return store.get_by_id(id)
        return cs.chained_get(self, get, artifact_id, put=_put_only_value)

    def contains_set(self, id):
        return cs.chained_contains(self, id, contains=_contains_set)

    def get_set_by_id(self, set_id):
        def get(store, id):
            return store.get_set_by_id(id)
        return cs.chained_get(self, get, set_id, put=_put_set)

    def get_set_by_name(self, set_name):
        def get(store, name):
            return store.get_set_by_name(name)
        return cs.chained_get(self, get, set_name, put=_put_set)

    def delete_set(self, id):
        return cs.chained_delete(self, id,
                                 contains=_contains_set, delete=_delete_set)

    def get_by_value_id(self, value_id):
        def get(store, id):
            return store.get_by_value_id(id)
        return cs.chained_get(self, get, value_id, put=_put_only_value)

    def get_value(self, artifact):
        for store in self.stores:
            try:
                return store.get_value(artifact)
            except KeyError:
                pass
        raise KeyError(artifact, self)

    def delete(self, id):
        return cs.chained_delete(self, id)


### ArtifactSet logic

def _set_op(operator, *sets, name=None):
    new_ids = t.reduce(operator, t.map(lambda s: s.artifact_ids, sets))
    return ArtifactSet(new_ids, name)


set_union = t.partial(_set_op, ops.or_)
set_difference = t.partial(_set_op, ops.sub)
set_intersection = t.partial(_set_op, ops.and_)

artifact_set_properties = ['id', 'artifact_ids', 'created_at', 'name']
class ArtifactSet(namedtuple('ArtifactSet', artifact_set_properties)):

    def __new__(cls, artifact_ids, name=None, created_at=None, id=None):
        artifact_ids = t.map(_artifact_id, artifact_ids)
        ids = frozenset(artifact_ids)
        if id:
            set_id = id
        else:
            set_id = hash(ids)
        created_at = created_at if created_at else datetime.utcnow()
        return super(ArtifactSet, cls).__new__(cls, set_id, ids, created_at, name)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.id == other.id
        return NotImplemented

    def artifacts_named(self, artifact_name):
        pass
        # return all artifacts found wrapped in list

    def add(self, artifact_or_id, name=None):
        artifact_id = _artifact_id(artifact_or_id)
        return ArtifactSet(self.artifact_ids | {artifact_id}, name)

    def remove(self, artifact_or_id, name=None):
        artifact_id = _artifact_id(artifact_or_id)
        return ArtifactSet(self.artifact_ids - {artifact_id}, name)

    def union(self, *sets, name=None):
        return set_union(self, *sets, name=name)

    def __or__(self, other_set):
        return set_union(self, other_set)

    def difference(self, *sets, name=None, repo=None):
        return set_difference(self, *sets, name=name)

    def __sub__(self, other_set):
        return set_difference(self, other_set)

    def intersection(self, *sets, name=None):
        return set_intersection(self, *sets, name=name)

    def __and__(self, other_set):
        return set_intersection(self, other_set)

    def rename(self, name):
        return self._replace(name=name)

    def put(self, repo=None):
        repo = repo if repo else get_default_repo()
        return repo.put_set(self)

    def proxy_dict(self, repo=None):
        return lazy_proxy_dict(self.artifact_ids)


def save_artifact(f, artifact_ids):

    def wrapped(*args, **kargs):
        artifact = f(*args, **kargs)
        artifact_ids.add(artifact.id)
        return artifact

    return wrapped


class RepoSpy(wrapt.ObjectProxy):
    def __init__(self, repo):
        super(RepoSpy, self).__init__(repo)
        self.artifact_ids = set()
        self.put = save_artifact(repo.put,
                                 self.artifact_ids)
        self.get_by_id = save_artifact(repo.get_by_id,
                                       self.artifact_ids)
        self.get_by_value_id = save_artifact(repo.get_by_value_id,
                                             self.artifact_ids)


@contextmanager
def capture_set(name=None, initial_set=None):
    if initial_set:
        initial = set(t.map(_artifact_id, initial_set))
    else:
        initial = set()

    repo = get_default_repo()
    spy = RepoSpy(repo)
    with using_repo(spy):
        result = []
        yield result
        artifact_ids = spy.artifact_ids | initial
    result.append(ArtifactSet(artifact_ids, name=name).put(repo))


def coerce_to_artifact(artifact_or_id, repo=None):
    repo = repo if repo else get_default_repo()
    if isinstance(artifact_or_id, string_type):
        return repo.get_by_id(artifact_or_id)
    if isinstance(artifact_or_id, Artifact):
        return artifact_or_id
    if is_proxy(artifact_or_id):
        return artifact_or_id.artifact
    raise ValueError('Was unable to coerce object into an Artifact: {}'
                     .format(artifact_or_id))

def coerce_to_artifacts(artifact_or_ids, repo=None):
    repo = repo if repo else get_default_repo()
    #TODO: bring this back when/if batch_get_by_id is added to chained repo
    # if all(isinstance(a, string_type) for a in artifact_or_ids):
    #     return repo.batch_get_by_id(artifact_or_ids)
    return [coerce_to_artifact(a, repo) for a in artifact_or_ids]


class lazy_dict(object):
    def __init__(self, thunks):
        self.thunks = thunks
        self.realized = {}

    def __getstate__(self):
        return self.thunks

    def __setstate__(self, thunks):
        self.__init__(thunks)

    def __getitem__(self, key):
        if key in self.thunks:
            if key not in self.realized:
                self.realized[key] = self.thunks[key]()
            return self.realized[key]
        else:
            raise KeyError(key, self)

    def __setitem__(self, key, value):
        self.thunks[key] = lambda: value
        self.realized[key] = value

    def __delitem__(self, key):
        if key in self.thunks:
            del self.thunks[key]
            if key in self.realized:
                del self.realized[key]
        else:
            KeyError(key, self)

    def __contains__(self, key):
        return key in self.thunks

    def items(self):
        return ((key, self[key]) for key in self.thunks.keys())

    def keys(self):
        return self.thunks.keys()

    def values(self):
        return (self[key] for key in self.thunks.keys())

    def __repr__(self):
        return "lazy_dict({})".format(
            t.merge(t.valmap(lambda _: "...", self.thunks), self.realized))

def lazy_proxy_dict(artifacts_or_ids):
    artifacts = coerce_to_artifacts(artifacts_or_ids)
    names = [a.name for a in artifacts]
    if not t.isdistinct(names):
        multi = t.thread_last(names,
                              t.frequencies,
                              (t.valfilter, lambda x: x > 1))
        raise ValueError("""Only artifacts with distinct names can be used in a lazy_proxy_dict.
-Offending names: {}""".format(multi))
    lambdas = {a.name: (lambda a: lambda: a.proxy())(a) for a in artifacts}

    return lazy_dict(lambdas)



import builtins
import inspect
import sys

PY3 = sys.version_info[0] == 3
PY2 = sys.version_info[0] == 2


if PY3:
    getargspec = inspect.getfullargspec
else:
    getargspec = inspect.getargspec

if hasattr(builtins, 'basestring'):
    string_type = basestring
else:
    string_type = str



import contextlib
import os
import os.path
import joblib as jl
import toolz as t
import shutil
from s3fs import S3FileSystem
import tempfile
from joblib.disk import mkdirp, rm_subdirs
from .serializers import DEFAULT_VALUE_SERIALIZER, DEFAULT_INPUT_SERIALIZER
from . import _commonstore as cs


class BaseBlobStore(object):
    def __init__(self, read=True, write=True, read_through_write=True,
                 delete=False, on_duplicate_key='skip'):
        self._read = read
        self._write = write
        self._read_through_write = read_through_write
        self._delete = delete
        self._on_duplicate_key = on_duplicate_key

        valid_on_duplicate_keys = {'skip', 'overwrite', 'check_collision', 'raise'}
        if self._on_duplicate_key not in valid_on_duplicate_keys:
            msg = "on_duplicate_key must be one of {}".format(valid_on_duplicate_keys)
            raise RuntimeError(msg)

    def __getitem__(self, id, *args, **kargs):
        return self.get(id, *args, **kargs)

    def put(self, id, value, serializer=DEFAULT_VALUE_SERIALIZER, read_through=False):
        method = getattr(self, '_put_' + self._on_duplicate_key)
        return method(id, value, serializer, read_through)

    def _put_raise(self, id, value, serializer, read_through):
        cs.ensure_put(self, id, read_through)
        self._put_overwrite(id, value, serializer, read_through)

    def _put_skip(self, id, value, serializer, read_through):
        if id not in self:
            self._put_overwrite(id, value, serializer, read_through)

    def _put_check_collision(self, id, value, serializer, read_through):
        cs.ensure_put(self, id, read_through, check_contains=False)
        if id not in self:
            self._put_overwrite(id, value, serializer, read_through)
        else:
            self._check_collision(id, value, serializer)

    # TODO: Right now our only thought is that this can be
    # checked by using an alternate hash, this will require
    # deserializing the old value and running the hash algorithm
    # with an alternate hash
    def _check_collision(self, id, value, serializer):
        raise NotImplemented()


class MemoryStore(BaseBlobStore):
    def __init__(self, values=None, read=True, write=True, read_through_write=True,
                 delete=True, on_duplicate_key='skip'):
        super(MemoryStore, self).__init__(
            read=read, write=write, read_through_write=read_through_write,
            delete=delete, on_duplicate_key=on_duplicate_key)
        if values is None:
            self.values = {}
        else:
            self.values = values

    def __contains__(self, id):
        cs.ensure_contains(self)
        return id in self.values

    def _put_overwrite(self, id, value, serializer, read_through):
        cs.ensure_put(self, id, read_through, check_contains=False)
        self.values[id] = value

    def get(self, id, serialzier=None, **_kargs):
        cs.ensure_read(self)
        cs.ensure_present(self, id)
        return self.values[id]

    def delete(self, id):
        cs.ensure_delete(self, id)
        del self.values[id]


@contextlib.contextmanager
def _temp_filename():
    try:
        temp = tempfile.NamedTemporaryFile('wb', delete=False)
        temp.close()
        yield temp.name
    finally:
        if os.path.isfile(temp.name):
            os.remove(temp.name)


@contextlib.contextmanager
def _atomic_write(filename):
    with _temp_filename() as temp:
        yield temp
        shutil.move(temp, filename)


def _abspath(path):
    return os.path.abspath(os.path.expanduser(path))


class DiskStore(BaseBlobStore):
    def __init__(self, cachedir, read=True, write=True, read_through_write=True,
                 delete=False, on_duplicate_key='skip'):
        super(DiskStore, self).__init__(
            read=read, write=write, read_through_write=read_through_write,
            delete=delete, on_duplicate_key=on_duplicate_key)
        self.cachedir = _abspath(cachedir)
        mkdirp(self.cachedir)

    def _filename(self, id):
        return os.path.join(self.cachedir, id)

    def __contains__(self, id):
        cs.ensure_contains(self)
        return os.path.isfile(self._filename(id))

    def _put_overwrite(self, id, value, serializer, read_through):
        cs.ensure_put(self, id, read_through, check_contains=False)
        with _atomic_write(self._filename(id)) as temp:
            serializer.dump(value, temp)

    def get(self, id, serializer=DEFAULT_VALUE_SERIALIZER, **_kargs):
        cs.ensure_read(self)
        cs.ensure_present(self, id)
        return serializer.load(self._filename(id))

    def delete(self, id):
        cs.ensure_delete(self, id)
        os.remove(self._filename(id))


class S3Store(BaseBlobStore):
    def __init__(self, cachedir, basepath, s3_config=None, s3fs=None,
                 read=True, write=True, read_through_write=True,
                 delete=False, on_duplicate_key='skip', cleanup_cachedir=False):
        super(S3Store, self).__init__(
            read=read, write=write, read_through_write=read_through_write,
            delete=delete, on_duplicate_key=on_duplicate_key)
        if s3fs:
            self.s3fs = s3fs
        elif s3_config is not None:
            self.s3fs = S3FileSystem(**s3_config)
        else:
            raise ValueError("You must provide either s3_config or s3fs for a S3Store")

        self.cachedir = _abspath(cachedir)
        self.basepath = basepath
        self.cleanup_cachedir = cleanup_cachedir
        mkdirp(self.cachedir)

    def __del__(self):
        if self.cleanup_cachedir:
            shutil.rmtree(self.cachedir)

    def _filename(self, id):
        return os.path.join(self.cachedir, id)

    def _path(self, id):
        return os.path.join(self.basepath, id)

    def __contains__(self, id):
        cs.ensure_contains(self)
        # maybe cache bucket listing if not too big?
        return self.s3fs.exists(self._path(id))

    def _put_overwrite(self, id, value, serializer, read_through):
        cs.ensure_put(self, id, read_through, check_contains=False)
        filename = self._filename(id)
        # not already saved by DiskStore?
        if not os.path.isfile(filename):
            with _atomic_write(filename) as temp:
                serializer.dump(value, temp)
        self.s3fs.put(filename, self._path(id))

    def get(self, id, serializer=DEFAULT_VALUE_SERIALIZER, **_kargs):
        cs.ensure_read(self)
        cs.ensure_present(self, id)
        filename = self._filename(id)
        if not os.path.exists(filename):
            with _atomic_write(filename) as temp:
                self.s3fs.get(self._path(id), temp)
        return serializer.load(filename)

    def delete(self, id):
        cs.ensure_delete(self, id)
        self.s3fs.rm(self._path(id))


class ChainedStore(object):
    def __init__(self, stores):
        self.stores = stores

    def __contains__(self, id):
        return cs.chained_contains(self, id)

    def _filename(self, id):
        if id in self:
            stores = [s for s in self.stores
                      if s._read and hasattr(s, '_filename')]
            if stores:
                return stores[0]._filename(id)
            else:
                raise Exception("You do not have a disk-based store setup.")

    def put(self, id, value, serializer=DEFAULT_VALUE_SERIALIZER):
        return cs.chained_put(self, id, value,
                              serializer=serializer)

    def get(self, id, serializer=DEFAULT_VALUE_SERIALIZER, **kargs):
        def get(store, id):
            return store.get(id, serializer=serializer, **kargs)
        return cs.chained_get(self, get, id)

    def __getitem__(self, id, **kargs):
        return self.get(id, **kargs)

    def delete(self, id):
        return cs.chained_delete(self, id)
